/**
 * @file
 * Contact matcher.
 *
 * Copyright (c) Microsoft Corporation. All rights reserved.
 * Licensed under the MIT License.
 */
import { MatcherConfig } from "./matcherconfig";
/**
 * Fields made available from the user defined Contact object for pronunciation and distance functions.
 *
 * @export
 * @interface ContactFields
 */
export interface ContactFields {
    /**
     * The name of the contact.
     *
     * @type {string}
     * @memberof ContactFields
     */
    name?: string;
    /**
     * The aliases the contact also goes by.
     *
     * @type {string[]}
     * @memberof ContactFields
     */
    aliases?: string[];
}
/**
 * Configurations to tweak the accuracy of the contact matcher.
 *
 * @export
 * @class ContactMatcherConfig
 */
export declare class ContactMatcherConfig extends MatcherConfig {
    /**
     *Creates an instance of ContactMatcherConfig.
     * @param {*} [{
     *         phoneticWeightPercentage = 0.7, Between 0 and 1. Weighting trade-off between the phonetic
     *  distance and the lexical distance scores. 1 meaning 100% phonetic score and 0% lexical score.
     *         maxReturns = 4, The maximum number of contacts the matcher can return.
     *         findThreshold = 0.35, The maximum distance to a match. Normalized to 0 for exact match, 1 for nothing matches.
     *  Can be >1 if the lengths do not match.
     *         maxDistanceMarginReturns = 0.02, Candidate cutoff given by
     *  Math.max({best matched distance} * bestDistanceMultiplier, maxDistanceMarginReturns).
     *         bestDistanceMultiplier = 1.1, Candidate cutoff given by
     *  Math.max({best matched distance} * bestDistanceMultiplier, maxDistanceMarginReturns).
     *  }={}]
     * @memberof ContactMatcherConfig
     */
    constructor({ phoneticWeightPercentage, maxReturns, findThreshold, maxDistanceMarginReturns, bestDistanceMultiplier }?: {
        phoneticWeightPercentage?: number | undefined;
        maxReturns?: number | undefined;
        findThreshold?: number | undefined;
        maxDistanceMarginReturns?: number | undefined;
        bestDistanceMultiplier?: number | undefined;
    });
}
/**
 * A fuzzy matcher that uses domain knowledge about contacts and sets up a simpler API.
 *
 * @export
 * @class EnContactMatcher
 * @template Contact The type of the contact object.
 */
export declare class EnContactMatcher<Contact> {
    readonly config: MatcherConfig;
    private static readonly tokenizer;
    private static readonly preprocessor;
    private readonly nameFuzzyMatcher;
    private readonly aliasFuzzyMatcher;
    private readonly nameMaxWindowSize;
    private readonly aliasMaxWindowSize;
    /**
     * Creates an instance of EnContactMatcher.
     *
     * @param {Contact[]} contacts The contacts to match against.
     * @param {(contact: Contact) => ContactFields} [extractContactFields=(contact: Contact): ContactFields => contact]
     *  Transform function from the user provided __Contact__ object to __ContactFields__, an intermediate representation used by the matcher.
     *  Defaults to the identity transform.
     * @param {MatcherConfig} [config=new ContactMatcherConfig()] The matcher's configuration object.
     * @memberof EnContactMatcher
     */
    constructor(contacts: Contact[], extractContactFields?: (contact: Contact) => ContactFields, config?: MatcherConfig);
    /**
     * Find a contact.
     *
     * @param {string} query The search query.
     * @returns {Contact[]} The matched contacts.
     * @memberof EnContactMatcher
     */
    find(query: string): Contact[];
    /**
     * Find a contact by only searching over their names.
     *
     * @param {string} name The name to search for.
     * @returns {Contact[]} The matched contacts.
     * @memberof EnContactMatcher
     */
    findByName(name: string): Contact[];
    /**
     * Find a contact by only searching over their aliases.
     *
     * @param {string} alias The alias to search for.
     * @returns {Contact[]} The matched contacts.
     * @memberof EnContactMatcher
     */
    findByAlias(alias: string): Contact[];
    private addNameVariations;
    private merge;
    private selectMatches;
}
