/**
 * @file
 * Place matcher.
 *
 * Copyright (c) Microsoft Corporation. All rights reserved.
 * Licensed under the MIT License.
 */
import { MatcherConfig } from "./matcherconfig";
/**
 * Fields made available from the user defined Place object for pronunciation and distance functions.
 *
 * @export
 * @interface PlaceFields
 */
export interface PlaceFields {
    /**
     * The name of the place.
     *
     * @type {string}
     * @memberof PlaceFields
     */
    name?: string;
    /**
     * The address of the place.
     *
     * @type {string}
     * @memberof PlaceFields
     */
    address?: string;
    /**
     * The tags/categories defining the place.
     *
     * @type {string[]}
     * @memberof PlaceFields
     */
    types?: string[];
}
/**
 * Configurations to tweak the accuracy of the place matcher.
 *
 * @export
 * @class PlaceMatcherConfig
 */
export declare class PlaceMatcherConfig extends MatcherConfig {
    /**
     *Creates an instance of PlaceMatcherConfig.
     * @param {*} [{
     *         phoneticWeightPercentage = 0.7, Between 0 and 1. Weighting trade-off between the phonetic
     *  distance and the lexical distance scores. 1 meaning 100% phonetic score and 0% lexical score.
     *         maxReturns = 8, The maximum number of places the matcher can return.
     *         findThreshold = 0.35, The maximum distance to a match. Normalized to 0 for exact match, 1 for nothing matches.
     *  Can be >1 if the lengths do not match.
     *         maxDistanceMarginReturns = 0.02, Candidate cutoff given by
     *  Math.max({best matched distance} * bestDistanceMultiplier, maxDistanceMarginReturns).
     *         bestDistanceMultiplier = 1.1,
     * }={}]
     * @memberof PlaceMatcherConfig
     */
    constructor({ phoneticWeightPercentage, maxReturns, findThreshold, maxDistanceMarginReturns, bestDistanceMultiplier }?: {
        phoneticWeightPercentage?: number | undefined;
        maxReturns?: number | undefined;
        findThreshold?: number | undefined;
        maxDistanceMarginReturns?: number | undefined;
        bestDistanceMultiplier?: number | undefined;
    });
}
/**
 * A fuzzy matcher that uses domain knowledge about places and sets up a simpler API.
 *
 * @export
 * @class EnPlaceMatcher
 * @template Place The type of the place object.
 */
export declare class EnPlaceMatcher<Place> {
    readonly config: MatcherConfig;
    private static readonly tokenizer;
    private static readonly preprocessor;
    private readonly fuzzyMatcher;
    private readonly maxWindowSize;
    /**
     * Creates an instance of EnPlaceMatcher.
     *
     * @param {Place[]} places The places to match against.
     * @param {(place: Place) => PlaceFields} [extractPlaceFields=(place: Place): PlaceFields => place]
     *  Transform function from the user provided __Place__ object to __PlaceFields__, an intermediate representation used by the matcher.
     *  Defaults to the identity transform.
     * @param {MatcherConfig} [config=new PlaceMatcherConfig()] The matcher's configuration object.
     * @memberof EnPlaceMatcher
     */
    constructor(places: Place[], extractPlaceFields?: (place: Place) => PlaceFields, config?: MatcherConfig);
    /**
     * Find a place.
     *
     * @param {string} query The search query.
     * @returns {Place[]} The matched places.
     * @memberof EnPlaceMatcher
     */
    find(query: string): Place[];
    private addNameVariations;
    private selectMatches;
}
