/**
 * @file
 * Pre-processors.
 *
 * Copyright (c) Microsoft Corporation. All rights reserved.
 * Licensed under the MIT License.
 */
/**
 * A Pre-processor. To transfor a string before any classification or understanding is known about it.
 *
 * @export
 * @interface PreProcessor
 */
export interface PreProcessor {
    /**
     * Function to preform the pre-processing.
     *
     * @param {string} query The string to pre-process.
     * @returns {string} The pre-processed string.
     * @memberof PreProcessor
     */
    preProcess(query: string): string;
}
/**
 * Pre-processes by appling a list of rules sequentially. First rules added are applied first.
 *
 * @export
 * @class ChainedRuleBasedPreProcessor
 * @implements {PreProcessor}
 */
export declare class ChainedRuleBasedPreProcessor implements PreProcessor {
    private readonly rules;
    /**
     * Pre-process a string.
     *
     * @param {string} query The string to pre-process.
     * @returns {string} The pre-processed string.
     * @memberof ChainedRuleBasedPreProcessor
     */
    preProcess(query: string): string;
    /**
     * Add a new rule.
     *
     * @param {RegExp} pattern The pattern to find that will be replaced.
     * @param {string} replacement The replacement string if the pattern is matched.
     * @memberof ChainedRuleBasedPreProcessor
     */
    add(pattern: RegExp, replacement: string): void;
}
/**
 * English Pre-processor.
 *
 * @export
 * @class EnPreProcessor
 * @implements {PreProcessor}
 */
export declare class EnPreProcessor implements PreProcessor {
    private readonly unicode;
    private readonly caseFold;
    protected readonly rules: ChainedRuleBasedPreProcessor;
    private readonly whitespace;
    private static STOP_WORDS;
    /**
     * Creates an instance of EnPreProcessor.
     *
     * @memberof EnPreProcessor
     */
    constructor();
    /**
     * Pre-process a string.
     *
     * @param {string} query The string to pre-process.
     * @returns {string} The pre-processed string.
     * @memberof EnPreProcessor
     */
    preProcess(query: string): string;
}
/**
 * English Pre-processor with specific rules for places.
 *
 * @export
 * @class EnPlacesPreProcessor
 * @extends {EnPreProcessor}
 */
export declare class EnPlacesPreProcessor extends EnPreProcessor {
    /**
     * Creates an instance of EnPlacesPreProcessor.
     *
     * @memberof EnPlacesPreProcessor
     */
    constructor();
}
