"use strict";
/**
 * @file
 * Pre-processors.
 *
 * Copyright (c) Microsoft Corporation. All rights reserved.
 * Licensed under the MIT License.
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.EnPlacesPreProcessor = exports.EnPreProcessor = exports.ChainedRuleBasedPreProcessor = void 0;
const xregexp_1 = __importDefault(require("xregexp"));
class UnicodePreProcessor {
    preProcess(query) {
        return query.normalize("NFKC");
    }
}
class CaseFoldingPreProcessor {
    preProcess(query) {
        // Unicode-aware only are recent versions of NodeJS.
        return query.toLowerCase();
    }
}
/**
 * Pre-processes by appling a list of rules sequentially. First rules added are applied first.
 *
 * @export
 * @class ChainedRuleBasedPreProcessor
 * @implements {PreProcessor}
 */
class ChainedRuleBasedPreProcessor {
    constructor() {
        this.rules = [];
    }
    /**
     * Pre-process a string.
     *
     * @param {string} query The string to pre-process.
     * @returns {string} The pre-processed string.
     * @memberof ChainedRuleBasedPreProcessor
     */
    preProcess(query) {
        let result = query;
        for (const rule of this.rules) {
            result = result.replace(rule.pattern, rule.replacement);
        }
        return result;
    }
    /**
     * Add a new rule.
     *
     * @param {RegExp} pattern The pattern to find that will be replaced.
     * @param {string} replacement The replacement string if the pattern is matched.
     * @memberof ChainedRuleBasedPreProcessor
     */
    add(pattern, replacement) {
        this.rules.push({ pattern, replacement });
    }
}
exports.ChainedRuleBasedPreProcessor = ChainedRuleBasedPreProcessor;
class WhiteSpacePreProcessor {
    constructor() {
        this.pattern = /\s{2,}/g;
    }
    preProcess(query) {
        return query.trim().replace(this.pattern, " ");
    }
}
/**
 * English Pre-processor.
 *
 * @export
 * @class EnPreProcessor
 * @implements {PreProcessor}
 */
class EnPreProcessor {
    /**
     * Creates an instance of EnPreProcessor.
     *
     * @memberof EnPreProcessor
     */
    constructor() {
        // TODO this belongs in native code to provide functionality cross language/platform. Will probably have to use libicu in some way.
        this.unicode = new UnicodePreProcessor();
        this.caseFold = new CaseFoldingPreProcessor();
        this.rules = new ChainedRuleBasedPreProcessor();
        this.whitespace = new WhiteSpacePreProcessor();
        // remove stop words
        this.rules.add(xregexp_1.default(`\\b(${EnPreProcessor.STOP_WORDS})\\b ?`, "g"), "");
        this.rules.add(xregexp_1.default(` ?\\b(${EnPreProcessor.STOP_WORDS})\\b`, "g"), "");
        // clear punctuation
        this.rules.add(xregexp_1.default("[\\p{P}\\p{S}]+", "g"), " ");
    }
    /**
     * Pre-process a string.
     *
     * @param {string} query The string to pre-process.
     * @returns {string} The pre-processed string.
     * @memberof EnPreProcessor
     */
    preProcess(query) {
        let result = query;
        result = this.unicode.preProcess(result);
        result = this.caseFold.preProcess(result);
        result = this.rules.preProcess(result);
        result = this.whitespace.preProcess(result);
        return result;
    }
}
exports.EnPreProcessor = EnPreProcessor;
EnPreProcessor.STOP_WORDS = ["a", "an", "at", "by", "el", "i", "in", "la", "las", "los", "my",
    "of", "on", "san", "santa", "some", "the", "with", "you"].join("|");
/**
 * English Pre-processor with specific rules for places.
 *
 * @export
 * @class EnPlacesPreProcessor
 * @extends {EnPreProcessor}
 */
class EnPlacesPreProcessor extends EnPreProcessor {
    /**
     * Creates an instance of EnPlacesPreProcessor.
     *
     * @memberof EnPlacesPreProcessor
     */
    constructor() {
        super();
        // Cardinal Directions
        this.rules.add(xregexp_1.default("\\be\\b", "g"), "east");
        this.rules.add(xregexp_1.default("\\bn\\b", "g"), "north");
        this.rules.add(xregexp_1.default("\\bs\\b", "g"), "south");
        this.rules.add(xregexp_1.default("\\bw\\b", "g"), "west");
        this.rules.add(xregexp_1.default("\\bne\\b", "g"), "north east");
        this.rules.add(xregexp_1.default("\\bnw\\b", "g"), "north west");
        this.rules.add(xregexp_1.default("\\bse\\b", "g"), "south east");
        this.rules.add(xregexp_1.default("\\bsw\\b", "g"), "south west");
        // Address Abbreviations
        // Word boundary doesn't work after the "." so we need look-ahead.
        this.rules.add(xregexp_1.default("\\baly\.?(?=[\\s\\p{P}\\p{S}]|$)", "g"), "alley");
        this.rules.add(xregexp_1.default("\\bave?\.?(?=[\\s\\p{P}\\p{S}]|$)", "g"), "avenue");
        this.rules.add(xregexp_1.default("\\bblvd\.?(?=[\\s\\p{P}\\p{S}]|$)", "g"), "boulevard");
        this.rules.add(xregexp_1.default("\\bbnd\.?(?=[\\s\\p{P}\\p{S}]|$)", "g"), "bend");
        this.rules.add(xregexp_1.default("\\bcres\.?(?=[\\s\\p{P}\\p{S}]|$)", "g"), "crescent");
        this.rules.add(xregexp_1.default("\\bcir\.?(?=[\\s\\p{P}\\p{S}]|$)", "g"), "circle");
        this.rules.add(xregexp_1.default("\\bct\.?(?=[\\s\\p{P}\\p{S}]|$)", "g"), "court");
        this.rules.add(xregexp_1.default("\\bdr\.?(?=[\\s\\p{P}\\p{S}]|$)", "g"), "drive");
        this.rules.add(xregexp_1.default("\\best\.?(?=[\\s\\p{P}\\p{S}]|$)", "g"), "estate");
        this.rules.add(xregexp_1.default("\\bln\.?(?=[\\s\\p{P}\\p{S}]|$)", "g"), "lane");
        this.rules.add(xregexp_1.default("\\bpkwy\.?(?=[\\s\\p{P}\\p{S}]|$)", "g"), "parkway");
        this.rules.add(xregexp_1.default("\\bpl\.?(?=[\\s\\p{P}\\p{S}]|$)", "g"), "place");
        this.rules.add(xregexp_1.default("\\brd\.?(?=[\\s\\p{P}\\p{S}]|$)", "g"), "road");
        // Assume "st" at the beginning is for "saint".
        this.rules.add(xregexp_1.default("^st\.?(?=[\\s\\p{P}\\p{S}]|$)", "g"), "saint");
        // If "st" does not occur at the start of the string, then we cannot known if it is for "saint" or "street".
        this.rules.add(xregexp_1.default("\\bst\.?(?=[\\s\\p{P}\\p{S}]|$)", "g"), "street");
        this.rules.add(xregexp_1.default("\\bxing\.?(?=[\\s\\p{P}\\p{S}]|$)", "g"), "crossing");
    }
}
exports.EnPlacesPreProcessor = EnPlacesPreProcessor;
