/**
 * @file
 * Tokenizers.
 *
 * Copyright (c) Microsoft Corporation. All rights reserved.
 * Licensed under the MIT License.
 */
/**
 * An Interval holds the first and last index bounds.
 *
 * @export
 * @class Interval
 */
export declare class Interval {
    readonly first: number;
    readonly last: number;
    /**
     * Creates an instance of Interval.
     *
     * @param {number} first Starting index (inclusive).
     * @param {number} last Ending index (exclusive).
     * @memberof Interval
     */
    constructor(first: number, last: number);
    /**
     * The length of the token.
     *
     * @returns {number} The length.
     * @memberof Interval
     */
    length(): number;
}
/**
 * The substring token of the original string with its interval location.
 *
 * @export
 * @class Token
 */
export declare class Token {
    readonly value: string;
    readonly interval: Interval;
    /**
     * Creates an instance of Token.
     *
     * @param {string} value The substring.
     * @param {Interval} interval The interval location.
     * @memberof Token
     */
    constructor(value: string, interval: Interval);
}
/**
 * Tokenizer interface for strings.
 *
 * @export
 * @interface Tokenizer
 */
export interface Tokenizer {
    /**
     * Tokenizes a string.
     *
     * @param {string} query The string to tokenize.
     * @returns {Token[]} The tokens.
     * @memberof Tokenizer
     */
    tokenize(query: string): Token[];
}
/**
 * Tokenizing base-class that will split on the given RegExp.
 *
 * @export
 * @abstract
 * @class SplittingTokenizer
 * @implements {Tokenizer}
 */
export declare abstract class SplittingTokenizer implements Tokenizer {
    private readonly pattern;
    /**
     * Creates an instance of SplittingTokenizer.
     *
     * @param {RegExp} pattern The pattern to split on.
     * @memberof SplittingTokenizer
     */
    constructor(pattern: RegExp);
    tokenize(query: string): Token[];
}
/**
 * Tokenizer that splits on whitespace.
 *
 * @export
 * @class WhitespaceTokenizer
 * @extends {SplittingTokenizer}
 */
export declare class WhitespaceTokenizer extends SplittingTokenizer {
    constructor();
}
