"use strict";
/**
 * @file
 * Tokenizers.
 *
 * Copyright (c) Microsoft Corporation. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.WhitespaceTokenizer = exports.SplittingTokenizer = exports.Token = exports.Interval = void 0;
/**
 * An Interval holds the first and last index bounds.
 *
 * @export
 * @class Interval
 */
class Interval {
    /**
     * Creates an instance of Interval.
     *
     * @param {number} first Starting index (inclusive).
     * @param {number} last Ending index (exclusive).
     * @memberof Interval
     */
    constructor(first, last) {
        this.first = first;
        this.last = last;
    }
    /**
     * The length of the token.
     *
     * @returns {number} The length.
     * @memberof Interval
     */
    length() {
        return this.last - this.first;
    }
}
exports.Interval = Interval;
/**
 * The substring token of the original string with its interval location.
 *
 * @export
 * @class Token
 */
class Token {
    /**
     * Creates an instance of Token.
     *
     * @param {string} value The substring.
     * @param {Interval} interval The interval location.
     * @memberof Token
     */
    constructor(value, interval) {
        this.value = value;
        this.interval = interval;
    }
}
exports.Token = Token;
/**
 * Tokenizing base-class that will split on the given RegExp.
 *
 * @export
 * @abstract
 * @class SplittingTokenizer
 * @implements {Tokenizer}
 */
class SplittingTokenizer {
    /**
     * Creates an instance of SplittingTokenizer.
     *
     * @param {RegExp} pattern The pattern to split on.
     * @memberof SplittingTokenizer
     */
    constructor(pattern) {
        this.pattern = pattern;
    }
    tokenize(query) {
        const result = [];
        let boundary = 0;
        let match;
        while ((match = this.pattern.exec(query)) !== null) {
            if (boundary < match.index) {
                const interval = new Interval(boundary, match.index);
                const token = new Token(query.substring(interval.first, interval.last), interval);
                result.push(token);
            }
            boundary = this.pattern.lastIndex;
        }
        if (boundary < query.length) {
            // Add the rest.
            const interval = new Interval(boundary, query.length);
            const token = new Token(query.substring(interval.first, interval.last), interval);
            result.push(token);
        }
        return result;
    }
}
exports.SplittingTokenizer = SplittingTokenizer;
/**
 * Tokenizer that splits on whitespace.
 *
 * @export
 * @class WhitespaceTokenizer
 * @extends {SplittingTokenizer}
 */
class WhitespaceTokenizer extends SplittingTokenizer {
    constructor() {
        super(/\s+/g);
    }
}
exports.WhitespaceTokenizer = WhitespaceTokenizer;
